@@ -30,16 +30,25 @@
    The GNU General Public License is contained in the file COPYING.
 */
 
+#include "pub_tool_basics.h"
+#include "pub_tool_hashtable.h"     // For mac_shared.h
+#include "pub_tool_libcbase.h"
+#include "pub_tool_libcassert.h"
+#include "pub_tool_libcprint.h"
+#include "pub_tool_profile.h"       // For mac_shared.h
+#include "pub_tool_tooliface.h"
+#include "pub_tool_threadstate.h"
+
 #include "mac_shared.h"
 #include "memcheck.h"
-//#include "vg_profile.c"
 
+#if 0
 
 /*------------------------------------------------------------*/
 /*--- Comparing and printing errors                        ---*/
 /*------------------------------------------------------------*/
 
-void SK_(pp_SkinError) ( Error* err )
+static void ac_pp_Error ( Error* err )
 {
    MAC_Error* err_extra = VG_(get_error_extra)(err);
 
@@ -66,7 +75,7 @@ void SK_(pp_SkinError) ( Error* err )
          break;
 
       default: 
-         MAC_(pp_shared_SkinError)(err);
+         MAC_(pp_shared_Error)(err);
          break;
    }
 }
@@ -75,7 +84,7 @@ void SK_(pp_SkinError) ( Error* err )
 /*--- Suppressions                                         ---*/
 /*------------------------------------------------------------*/
 
-Bool SK_(recognised_suppression) ( Char* name, Supp* su )
+static Bool ac_recognised_suppression ( Char* name, Supp* su )
 {
    return MAC_(shared_recognised_suppression)(name, su);
 }
@@ -102,7 +111,7 @@ Bool SK_(recognised_suppression) ( Char* name, Supp* su )
    All entries in the primary (top-level) map must point to a valid
    secondary (second-level) map.  Since most of the 4G of address
    space will not be in use -- ie, not mapped at all -- there is a
-   distinguished secondary map, which indicates `not addressible and
+   distinguished secondary map, which indicates 'not addressible and
    not valid' writeable for all bytes.  Entries in the primary map for
    which the entire 64k is not in use at all point at this
    distinguished map.
@@ -169,11 +178,11 @@ static void init_shadow_memory ( void )
    Int i, a;
 
    /* check construction of the distinguished secondaries */
-   sk_assert(VGM_BIT_INVALID == 1);
-   sk_assert(VGM_BIT_VALID == 0);
+   tl_assert(VGM_BIT_INVALID == 1);
+   tl_assert(VGM_BIT_VALID == 0);
 
    for(a = 0; a <= 1; a++)
-      sk_assert(distinguished_secondary_maps[DSM_IDX(a)].abits[0] == BIT_EXPAND(a));
+      tl_assert(distinguished_secondary_maps[DSM_IDX(a)].abits[0] == BIT_EXPAND(a));
 
    /* These entries gradually get overwritten as the used address
       space expands. */
@@ -196,7 +205,7 @@ static AcSecMap* alloc_secondary_map ( __attribute__ ((unused))
 				       const AcSecMap *prototype)
 {
    AcSecMap* map;
-   PROF_EVENT(10);
+   PROF_EVENT(10, "");
 
    map = (AcSecMap *)VG_(shadow_alloc)(sizeof(AcSecMap));
    VG_(memcpy)(map, prototype, sizeof(*map));
@@ -212,7 +221,7 @@ static __inline__ UChar get_abit ( Addr a )
 {
    AcSecMap* sm     = primary_map[PM_IDX(a)];
    UInt    sm_off = SM_OFF(a);
-   PROF_EVENT(20);
+   PROF_EVENT(20, "");
 #  if 0
       if (IS_DISTINGUISHED_SM(sm))
          VG_(message)(Vg_DebugMsg, 
@@ -226,7 +235,7 @@ static /* __inline__ */ void set_abit ( Addr a, UChar abit )
 {
    AcSecMap* sm;
    UInt    sm_off;
-   PROF_EVENT(22);
+   PROF_EVENT(22, "");
    ENSURE_MAPPABLE(a, "set_abit");
    sm     = primary_map[PM_IDX(a)];
    sm_off = SM_OFF(a);
@@ -244,9 +253,9 @@ static __inline__ UChar get_abits4_ALIGNED ( Addr a )
    AcSecMap* sm;
    UInt    sm_off;
    UChar   abits8;
-   PROF_EVENT(24);
+   PROF_EVENT(24, "");
 #  ifdef VG_DEBUG_MEMORY
-   sk_assert(IS_ALIGNED4_ADDR(a));
+   tl_assert(VG_IS_4_ALIGNED(a));
 #  endif
    sm     = primary_map[PM_IDX(a)];
    sm_off = SM_OFF(a);
@@ -269,7 +278,7 @@ void set_address_range_perms ( Addr a, SizeT len, UInt example_a_bit )
    UInt      sm_off;
    AcSecMap* sm;
 
-   PROF_EVENT(30);
+   PROF_EVENT(30, "");
 
    if (len == 0)
       return;
@@ -287,10 +296,10 @@ void set_address_range_perms ( Addr a, SizeT len, UInt example_a_bit )
       indicate bugs in our machinery.  30,000,000 is arbitrary, but so
       far all legitimate requests have fallen beneath that size. */
    /* 4 Mar 02: this is just stupid; get rid of it. */
-   /* sk_assert(len < 30000000); */
+   /* tl_assert(len < 30000000); */
 
    /* Check the permissions make sense. */
-   sk_assert(example_a_bit == VGM_BIT_VALID 
+   tl_assert(example_a_bit == VGM_BIT_VALID 
              || example_a_bit == VGM_BIT_INVALID);
 
    /* In order that we can charge through the address space at 8
@@ -300,7 +309,7 @@ void set_address_range_perms ( Addr a, SizeT len, UInt example_a_bit )
 #  ifdef VG_DEBUG_MEMORY
    /* Do it ... */
    while (True) {
-      PROF_EVENT(31);
+      PROF_EVENT(31, "");
       if (len == 0) break;
       set_abit ( a, example_a_bit );
       set_vbyte ( a, vbyte );
@@ -311,7 +320,7 @@ void set_address_range_perms ( Addr a, SizeT len, UInt example_a_bit )
 #  else
    /* Slowly do parts preceding 8-byte alignment. */
    while (True) {
-      PROF_EVENT(31);
+      PROF_EVENT(31, "");
       if (len == 0) break;
       if ((a % 8) == 0) break;
       set_abit ( a, example_a_bit );
@@ -323,11 +332,11 @@ void set_address_range_perms ( Addr a, SizeT len, UInt example_a_bit )
       VGP_POPCC(VgpSetMem);
       return;
    }
-   sk_assert((a % 8) == 0 && len > 0);
+   tl_assert((a % 8) == 0 && len > 0);
 
    /* Once aligned, go fast up to primary boundary. */
    for (; (a & SECONDARY_MASK) && len >= 8; a += 8, len -= 8) {
-      PROF_EVENT(32);
+      PROF_EVENT(32, "");
 
       /* If the primary is already pointing to a distinguished map
 	 with the same properties as we're trying to set, then leave
@@ -357,7 +366,7 @@ void set_address_range_perms ( Addr a, SizeT len, UInt example_a_bit )
 
    /* Now finished the remains. */
    for (; len >= 8; a += 8, len -= 8) {
-      PROF_EVENT(32);
+      PROF_EVENT(32, "");
 
       /* If the primary is already pointing to a distinguished map
 	 with the same properties as we're trying to set, then leave
@@ -370,9 +379,10 @@ void set_address_range_perms ( Addr a, SizeT len, UInt example_a_bit )
       sm->abits[sm_off >> 3] = abyte8;
    }
 
+
    /* Finish the upper fragment. */
    while (True) {
-      PROF_EVENT(33);
+      PROF_EVENT(33, "");
       if (len == 0) break;
       set_abit ( a, example_a_bit );
       a++;
@@ -380,10 +390,6 @@ void set_address_range_perms ( Addr a, SizeT len, UInt example_a_bit )
    }   
 #  endif
 
-   /* Check that zero page and highest page have not been written to
-      -- this could happen with buggy syscall wrappers.  Today
-      (2001-04-26) had precisely such a problem with __NR_setitimer. */
-   sk_assert(SK_(cheap_sanity_check)());
    VGP_POPCC(VgpSetMem);
 }
 
@@ -391,14 +397,14 @@ void set_address_range_perms ( Addr a, SizeT len, UInt example_a_bit )
 
 static void ac_make_noaccess ( Addr a, SizeT len )
 {
-   PROF_EVENT(35);
+   PROF_EVENT(35, "");
    DEBUG("ac_make_noaccess(%p, %x)\n", a, len);
    set_address_range_perms ( a, len, VGM_BIT_INVALID );
 }
 
 static void ac_make_accessible ( Addr a, SizeT len )
 {
-   PROF_EVENT(38);
+   PROF_EVENT(38, "");
    DEBUG("ac_make_accessible(%p, %x)\n", a, len);
    set_address_range_perms ( a, len, VGM_BIT_VALID );
 }
@@ -487,10 +493,10 @@ static void ac_copy_address_range_state ( Addr src, Addr dst, SizeT len )
 
    DEBUG("ac_copy_address_range_state\n");
 
-   PROF_EVENT(40);
+   PROF_EVENT(40, "");
    for (i = 0; i < len; i++) {
       UChar abit  = get_abit ( src+i );
-      PROF_EVENT(41);
+      PROF_EVENT(41, "");
       set_abit ( dst+i, abit );
    }
 }
@@ -505,9 +511,9 @@ Bool ac_check_accessible ( Addr a, SizeT len, Addr* bad_addr )
 {
    UInt  i;
    UChar abit;
-   PROF_EVENT(48);
+   PROF_EVENT(48, "");
    for (i = 0; i < len; i++) {
-      PROF_EVENT(49);
+      PROF_EVENT(49, "");
       abit = get_abit(a);
       if (abit == VGM_BIT_INVALID) {
          if (bad_addr != NULL) *bad_addr = a;
@@ -524,9 +530,9 @@ Bool ac_check_noaccess ( Addr a, SizeT len, Addr* bad_addr )
 {
    UInt  i;
    UChar abit;
-   PROF_EVENT(48);
+   PROF_EVENT(48, "");
    for (i = 0; i < len; i++) {
-      PROF_EVENT(49);
+      PROF_EVENT(49, "");
       abit = get_abit(a);
       if (abit == VGM_BIT_VALID) {
          if (bad_addr != NULL) *bad_addr = a;
@@ -545,10 +551,10 @@ static __inline__
 Bool ac_check_readable_asciiz ( Addr a, Addr* bad_addr )
 {
    UChar abit;
-   PROF_EVENT(46);
+   PROF_EVENT(46, "");
    DEBUG("ac_check_readable_asciiz\n");
    while (True) {
-      PROF_EVENT(47);
+      PROF_EVENT(47, "");
       abit  = get_abit(a);
       if (abit != VGM_BIT_VALID) {
          if (bad_addr != NULL) *bad_addr = a;
@@ -570,7 +576,7 @@ void ac_check_is_accessible ( CorePart part, ThreadId tid,
                               Char* s, Addr base, SizeT size, Bool isWrite )
 {
    Bool ok;
-   Addr bad_addr;
+   Addr bad_addr = 0;   // Initialise to shut gcc up
 
    VGP_PUSHCC(VgpCheckMem);
 
@@ -583,7 +589,7 @@ void ac_check_is_accessible ( CorePart part, ThreadId tid,
          break;
 
       case Vg_CoreSignal:
-         sk_assert(isWrite);     /* Should only happen with isWrite case */
+         tl_assert(isWrite);     /* Should only happen with isWrite case */
          /* fall through */
       case Vg_CorePThread:
          MAC_(record_core_mem_error)( tid, /*isUnaddr*/True, s );
@@ -592,12 +598,12 @@ void ac_check_is_accessible ( CorePart part, ThreadId tid,
       /* If we're being asked to jump to a silly address, record an error 
          message before potentially crashing the entire system. */
       case Vg_CoreTranslate:
-         sk_assert(!isWrite);    /* Should only happen with !isWrite case */
+         tl_assert(!isWrite);    /* Should only happen with !isWrite case */
          MAC_(record_jump_error)( tid, bad_addr );
          break;
 
       default:
-         VG_(skin_panic)("ac_check_is_accessible: unexpected CorePart");
+         VG_(tool_panic)("ac_check_is_accessible: unexpected CorePart");
       }
    }
 
@@ -623,11 +629,11 @@ void ac_check_is_readable_asciiz ( CorePart part, ThreadId tid,
                                    Char* s, Addr str )
 {
    Bool ok = True;
-   Addr bad_addr;
+   Addr bad_addr = 0;   // Initialise to shut gcc up
 
    VGP_PUSHCC(VgpCheckMem);
 
-   sk_assert(part == Vg_CoreSysCall);
+   tl_assert(part == Vg_CoreSysCall);
    ok = ac_check_readable_asciiz ( (Addr)str, &bad_addr );
    if (!ok) {
       MAC_(record_param_error) ( tid, bad_addr, /*IsReg*/False,
@@ -659,6 +665,12 @@ void ac_new_mem_mmap (Addr a, SizeT len, Bool rr, Bool ww, Bool xx)
    ac_make_accessible(a, len);
 }
 
+static
+void ac_post_mem_write(CorePart part, ThreadId tid, Addr a, SizeT len)
+{
+   ac_make_accessible(a, len);
+}
+
 
 /*------------------------------------------------------------*/
 /*--- Functions called directly from generated code.       ---*/
@@ -694,7 +706,8 @@ static __inline__ void ac_helperc_ACCESS4 ( Addr a, Bool isWrite )
    UChar   abits  = sm->abits[a_off];
    abits >>= (a & 4);
    abits &= 15;
-   PROF_EVENT(66);
+   PROF_EVENT(66, "");
+   // XXX: Should this be "if (!is_distinguished_sm(sm) && abits == VGM_NIBBLE_VALID)"?
    if (abits == VGM_NIBBLE_VALID) {
       /* Handle common case quickly: a is suitably aligned, is mapped,
          and is addressible.  So just return. */
@@ -714,7 +727,8 @@ static __inline__ void ac_helperc_ACCESS2 ( Addr a, Bool isWrite )
    UInt    sec_no = rotateRight16(a) & 0x1FFFF;
    AcSecMap* sm     = primary_map[sec_no];
    UInt    a_off  = (SM_OFF(a)) >> 3;
-   PROF_EVENT(67);
+   PROF_EVENT(67, "");
+   // XXX: Should this be "if (!is_distinguished_sm(sm) && sm->abits[a_off] == VGM_BYTE_VALID)"?
    if (sm->abits[a_off] == VGM_BYTE_VALID) {
       /* Handle common case quickly. */
       return;
@@ -733,7 +747,8 @@ static __inline__ void ac_helperc_ACCESS1 ( Addr a, Bool isWrite )
    UInt    sec_no = shiftRight16(a);
    AcSecMap* sm   = primary_map[sec_no];
    UInt    a_off  = (SM_OFF(a)) >> 3;
-   PROF_EVENT(68);
+   PROF_EVENT(68, "");
+   // XXX: Should this be "if (!is_distinguished_sm(sm) && sm->abits[a_off] == VGM_BYTE_VALID)"?
    if (sm->abits[a_off] == VGM_BYTE_VALID) {
       /* Handle common case quickly. */
       return;
@@ -744,35 +759,35 @@ static __inline__ void ac_helperc_ACCESS1 ( Addr a, Bool isWrite )
 #  endif
 }
 
-REGPARM(1)
-static void ac_helperc_LOAD4 ( Addr a )
+VG_REGPARM(1)
+static void ach_LOAD4 ( Addr a )
 {
    ac_helperc_ACCESS4 ( a, /*isWrite*/False );
 }
-REGPARM(1)
-static void ac_helperc_STORE4 ( Addr a )
+VG_REGPARM(1)
+static void ach_STORE4 ( Addr a )
 {
    ac_helperc_ACCESS4 ( a, /*isWrite*/True );
 }
 
-REGPARM(1)
-static void ac_helperc_LOAD2 ( Addr a )
+VG_REGPARM(1)
+static void ach_LOAD2 ( Addr a )
 {
    ac_helperc_ACCESS2 ( a, /*isWrite*/False );
 }
-REGPARM(1)
-static void ac_helperc_STORE2 ( Addr a )
+VG_REGPARM(1)
+static void ach_STORE2 ( Addr a )
 {
    ac_helperc_ACCESS2 ( a, /*isWrite*/True );
 }
 
-REGPARM(1)
-static void ac_helperc_LOAD1 ( Addr a )
+VG_REGPARM(1)
+static void ach_LOAD1 ( Addr a )
 {
    ac_helperc_ACCESS1 ( a, /*isWrite*/False );
 }
-REGPARM(1)
-static void ac_helperc_STORE1 ( Addr a )
+VG_REGPARM(1)
+static void ach_STORE1 ( Addr a )
 {
    ac_helperc_ACCESS1 ( a, /*isWrite*/True );
 }
@@ -787,7 +802,7 @@ static void ac_ACCESS4_SLOWLY ( Addr a, Bool isWrite )
 {
    Bool a0ok, a1ok, a2ok, a3ok;
 
-   PROF_EVENT(76);
+   PROF_EVENT(76, "");
 
    /* First establish independently the addressibility of the 4 bytes
       involved. */
@@ -812,7 +827,7 @@ static void ac_ACCESS4_SLOWLY ( Addr a, Bool isWrite )
    if (!MAC_(clo_partial_loads_ok) 
        || ((a & 3) != 0)
        || (!a0ok && !a1ok && !a2ok && !a3ok)) {
-      MAC_(record_address_error)( VG_(get_VCPU_tid)(), a, 4, isWrite );
+      MAC_(record_address_error)( VG_(get_running_tid)(), a, 4, isWrite );
       return;
    }
 
@@ -822,7 +837,7 @@ static void ac_ACCESS4_SLOWLY ( Addr a, Bool isWrite )
       (which is the default), and the address is 4-aligned.  
       If not, Case 2 will have applied.
    */
-   sk_assert(MAC_(clo_partial_loads_ok));
+   tl_assert(MAC_(clo_partial_loads_ok));
    {
       return;
    }
@@ -832,14 +847,14 @@ static void ac_ACCESS2_SLOWLY ( Addr a, Bool isWrite )
 {
    /* Check the address for validity. */
    Bool aerr = False;
-   PROF_EVENT(77);
+   PROF_EVENT(77, "");
 
    if (get_abit(a+0) != VGM_BIT_VALID) aerr = True;
    if (get_abit(a+1) != VGM_BIT_VALID) aerr = True;
 
    /* If an address error has happened, report it. */
    if (aerr) {
-      MAC_(record_address_error)( VG_(get_VCPU_tid)(), a, 2, isWrite );
+      MAC_(record_address_error)( VG_(get_running_tid)(), a, 2, isWrite );
    }
 }
 
@@ -847,13 +862,13 @@ static void ac_ACCESS1_SLOWLY ( Addr a, Bool isWrite)
 {
    /* Check the address for validity. */
    Bool aerr = False;
-   PROF_EVENT(78);
+   PROF_EVENT(78, "");
 
    if (get_abit(a+0) != VGM_BIT_VALID) aerr = True;
 
    /* If an address error has happened, report it. */
    if (aerr) {
-      MAC_(record_address_error)( VG_(get_VCPU_tid)(), a, 1, isWrite );
+      MAC_(record_address_error)( VG_(get_running_tid)(), a, 1, isWrite );
    }
 }
 
@@ -877,15 +892,15 @@ void ac_fpu_ACCESS_check ( Addr addr, SizeT size, Bool isWrite )
    UInt    sm_off, a_off;
    Addr    addr4;
 
-   PROF_EVENT(90);
+   PROF_EVENT(90, "");
 
 #  ifdef VG_DEBUG_MEMORY
    ac_fpu_ACCESS_check_SLOWLY ( addr, size, isWrite );
 #  else
 
    if (size == 4) {
-      if (!IS_ALIGNED4_ADDR(addr)) goto slow4;
-      PROF_EVENT(91);
+      if (!VG_IS_4_ALIGNED(addr)) goto slow4;
+      PROF_EVENT(91, "");
       /* Properly aligned. */
       sm     = primary_map[PM_IDX(addr)];
       sm_off = SM_OFF(addr);
@@ -899,8 +914,8 @@ void ac_fpu_ACCESS_check ( Addr addr, SizeT size, Bool isWrite )
    }
 
    if (size == 8) {
-      if (!IS_ALIGNED4_ADDR(addr)) goto slow8;
-      PROF_EVENT(92);
+      if (!VG_IS_4_ALIGNED(addr)) goto slow8;
+      PROF_EVENT(92, "");
       /* Properly aligned.  Do it in two halves. */
       addr4 = addr + 4;
       /* First half. */
@@ -925,34 +940,35 @@ void ac_fpu_ACCESS_check ( Addr addr, SizeT size, Bool isWrite )
    /* Can't be bothered to huff'n'puff to make these (allegedly) rare
       cases go quickly.  */
    if (size == 2) {
-      PROF_EVENT(93);
+      PROF_EVENT(93, "");
       ac_fpu_ACCESS_check_SLOWLY ( addr, 2, isWrite );
       return;
    }
 
    if (size == 16 || size == 10 || size == 28 || size == 108 || size == 512) {
-      PROF_EVENT(94);
+      PROF_EVENT(94, "");
       ac_fpu_ACCESS_check_SLOWLY ( addr, size, isWrite );
       return;
    }
 
    VG_(printf)("size is %d\n", size);
-   VG_(skin_panic)("fpu_ACCESS_check: unhandled size");
+   VG_(tool_panic)("fpu_ACCESS_check: unhandled size");
 #  endif
 }
 
-REGPARM(2)
-static void ac_fpu_READ_check ( Addr addr, SizeT size )
+VG_REGPARM(2)
+static void ach_LOADN ( Addr addr, SizeT size )
 {
    ac_fpu_ACCESS_check ( addr, size, /*isWrite*/False );
 }
 
-REGPARM(2)
-static void ac_fpu_WRITE_check ( Addr addr, SizeT size )
+VG_REGPARM(2)
+static void ach_STOREN ( Addr addr, SizeT size )
 {
    ac_fpu_ACCESS_check ( addr, size, /*isWrite*/True );
 }
 
+
 /* ---------------------------------------------------------------------
    Slow, general cases for FPU access checks.
    ------------------------------------------------------------------ */
@@ -961,15 +977,15 @@ void ac_fpu_ACCESS_check_SLOWLY ( Addr addr, SizeT size, Bool isWrite )
 {
    Int  i;
    Bool aerr = False;
-   PROF_EVENT(100);
+   PROF_EVENT(100, "");
    for (i = 0; i < size; i++) {
-      PROF_EVENT(101);
+      PROF_EVENT(101, "");
       if (get_abit(addr+i) != VGM_BIT_VALID)
          aerr = True;
    }
 
    if (aerr) {
-      MAC_(record_address_error)( VG_(get_VCPU_tid)(), addr, size, isWrite );
+      MAC_(record_address_error)( VG_(get_running_tid)(), addr, size, isWrite );
    }
 }
 
@@ -978,137 +994,156 @@ void ac_fpu_ACCESS_check_SLOWLY ( Addr addr, SizeT size, Bool isWrite )
 /*--- Our instrumenter                                     ---*/
 /*------------------------------------------------------------*/
 
-UCodeBlock* SK_(instrument)(UCodeBlock* cb_in, Addr orig_addr)
+static IRBB* ac_instrument(IRBB* bb_in, VexGuestLayout* layout, 
+                           IRType gWordTy, IRType hWordTy )
 {
-/* Use this rather than eg. -1 because it's a UInt. */
-#define INVALID_DATA_SIZE   999999
+   Int         i, hsz;
+   IRStmt*     st;
+   IRExpr*     data;
+   IRExpr*     aexpr;
+   IRExpr*     guard;
+   IRDirty*    di;
+   Bool        isLoad;
+   IRBB*       bb;
 
-   UCodeBlock* cb;
-   Int         i;
-   UInstr*     u_in;
-   Int         t_addr, t_size;
-   Addr        helper;
-
-   cb = VG_(setup_UCodeBlock)(cb_in);
-
-   for (i = 0; i < VG_(get_num_instrs)(cb_in); i++) {
-
-      t_addr = t_size = INVALID_TEMPREG;
-      u_in = VG_(get_instr)(cb_in, i);
-
-      switch (u_in->opcode) {
-         case NOP:  case LOCK:  case CALLM_E:  case CALLM_S:
-            break;
-
-         /* For memory-ref instrs, copy the data_addr into a temporary
-          * to be passed to the helper at the end of the instruction.
-          */
-         case LOAD:
-            switch (u_in->size) {
-               case 4:  helper = (Addr)ac_helperc_LOAD4; break;
-               case 2:  helper = (Addr)ac_helperc_LOAD2; break;
-               case 1:  helper = (Addr)ac_helperc_LOAD1; break;
-               default: VG_(skin_panic)
-                           ("addrcheck::SK_(instrument):LOAD");
-            }
-            uInstr1(cb, CCALL, 0, TempReg, u_in->val1);
-            uCCall (cb, helper, 1, 1, False );
-            VG_(copy_UInstr)(cb, u_in);
-            break;
-
-         case STORE:
-            switch (u_in->size) {
-               case 4:  helper = (Addr)ac_helperc_STORE4; break;
-               case 2:  helper = (Addr)ac_helperc_STORE2; break;
-               case 1:  helper = (Addr)ac_helperc_STORE1; break;
-               default: VG_(skin_panic)
-                           ("addrcheck::SK_(instrument):STORE");
-            }
-            uInstr1(cb, CCALL, 0, TempReg, u_in->val2);
-            uCCall (cb, helper, 1, 1, False );
-            VG_(copy_UInstr)(cb, u_in);
-            break;
-
-	 case SSE3ag_MemRd_RegWr:
-            sk_assert(u_in->size == 4 || u_in->size == 8);
-            helper = (Addr)ac_fpu_READ_check;
-	    goto do_Access_ARG1;
-         do_Access_ARG1:
-	    sk_assert(u_in->tag1 == TempReg);
-            t_addr = u_in->val1;
-            t_size = newTemp(cb);
-	    uInstr2(cb, MOV, 4, Literal, 0, TempReg, t_size);
-	    uLiteral(cb, u_in->size);
-            uInstr2(cb, CCALL, 0, TempReg, t_addr, TempReg, t_size);
-            uCCall(cb, helper, 2, 2, False );
-            VG_(copy_UInstr)(cb, u_in);
-            break;
-
-         case MMX2_MemRd:
-            sk_assert(u_in->size == 4 || u_in->size == 8);
-            helper = (Addr)ac_fpu_READ_check;
-	    goto do_Access_ARG2;
-         case MMX2_MemWr:
-            sk_assert(u_in->size == 4 || u_in->size == 8);
-            helper = (Addr)ac_fpu_WRITE_check;
-	    goto do_Access_ARG2;
-         case FPU_R:
-            helper = (Addr)ac_fpu_READ_check;
-            goto do_Access_ARG2;
-         case FPU_W:
-            helper = (Addr)ac_fpu_WRITE_check;
-            goto do_Access_ARG2;
-         do_Access_ARG2:
-	    sk_assert(u_in->tag2 == TempReg);
-            t_addr = u_in->val2;
-            t_size = newTemp(cb);
-	    uInstr2(cb, MOV, 4, Literal, 0, TempReg, t_size);
-	    uLiteral(cb, u_in->size);
-            uInstr2(cb, CCALL, 0, TempReg, t_addr, TempReg, t_size);
-            uCCall(cb, helper, 2, 2, False );
-            VG_(copy_UInstr)(cb, u_in);
-            break;
-
-         case MMX2a1_MemRd:
-         case SSE3a_MemRd:
-         case SSE2a_MemRd:
-         case SSE3a1_MemRd:
-         case SSE2a1_MemRd:
-            helper = (Addr)ac_fpu_READ_check;
-	    goto do_Access_ARG3;
-         case SSE2a_MemWr:
-	 case SSE3a_MemWr:
-            helper = (Addr)ac_fpu_WRITE_check;
-	    goto do_Access_ARG3;
-         do_Access_ARG3:
-	    sk_assert(u_in->size == 4 || u_in->size == 8
-                      || u_in->size == 16 || u_in->size == 512);
-            sk_assert(u_in->tag3 == TempReg);
-            t_addr = u_in->val3;
-            t_size = newTemp(cb);
-	    uInstr2(cb, MOV, 4, Literal, 0, TempReg, t_size);
-	    uLiteral(cb, u_in->size);
-            uInstr2(cb, CCALL, 0, TempReg, t_addr, TempReg, t_size);
-            uCCall(cb, helper, 2, 2, False );
-            VG_(copy_UInstr)(cb, u_in);
-            break;
-
-         case SSE3e1_RegRd:
-         case SSE3e_RegWr:
-         case SSE3g1_RegWr:
-         case SSE5:
-         case SSE3g_RegWr:
-         case SSE3e_RegRd:
-         case SSE4:
-         case SSE3:
-         default:
-            VG_(copy_UInstr)(cb, u_in);
-            break;
-      }
+   if (gWordTy != hWordTy) {
+      /* We don't currently support this case. */
+      VG_(tool_panic)("host/guest word size mismatch");
    }
 
-   VG_(free_UCodeBlock)(cb_in);
-   return cb;
+   /* Set up BB */
+   bb           = emptyIRBB();
+   bb->tyenv    = dopyIRTypeEnv(bb_in->tyenv);
+   bb->next     = dopyIRExpr(bb_in->next);
+   bb->jumpkind = bb_in->jumpkind;
+
+   /* No loads to consider in ->next. */
+   tl_assert(isIRAtom(bb_in->next));
+
+   for (i = 0; i <  bb_in->stmts_used; i++) {
+      st = bb_in->stmts[i];
+      tl_assert(st);
+
+      /* Examine each stmt in turn to figure out if it needs to be
+         preceded by a memory access check.  If so, collect up the
+         relevant pieces of information. */
+      hsz    = 0;
+      aexpr  = NULL;
+      guard  = NULL;
+      isLoad = True;
+
+      switch (st->tag) {
+
+         case Ist_Tmp:
+            data = st->Ist.Tmp.data;
+            if (data->tag == Iex_LDle) {
+               aexpr  = data->Iex.LDle.addr;
+               hsz    = sizeofIRType(data->Iex.LDle.ty);
+               isLoad = True;
+	    }
+	    break;
+
+         case Ist_STle:
+            data  = st->Ist.STle.data;
+            aexpr = st->Ist.STle.addr;
+            tl_assert(isIRAtom(data));
+            tl_assert(isIRAtom(aexpr));
+            hsz    = sizeofIRType(typeOfIRExpr(bb_in->tyenv, data));
+	    isLoad = False;
+            break;
+
+         case Ist_Put:
+            tl_assert(isIRAtom(st->Ist.Put.data));
+            break;
+
+         case Ist_PutI:
+            tl_assert(isIRAtom(st->Ist.PutI.ix));
+            tl_assert(isIRAtom(st->Ist.PutI.data));
+            break;
+
+         case Ist_Exit:
+            tl_assert(isIRAtom(st->Ist.Exit.guard));
+            break;
+
+         case Ist_Dirty:
+            if (st->Ist.Dirty.details->mFx != Ifx_None) {
+               /* We classify Ifx_Modify as a load. */
+               isLoad = st->Ist.Dirty.details->mFx != Ifx_Write;
+               hsz    = st->Ist.Dirty.details->mSize;
+               aexpr  = st->Ist.Dirty.details->mAddr;
+               guard  = st->Ist.Dirty.details->guard;
+               tl_assert(isIRAtom(aexpr));
+             }
+             break;
+
+         case Ist_NoOp:
+         case Ist_IMark:
+         case Ist_MFence:
+            break;
+
+         default:
+            VG_(printf)("\n");
+            ppIRStmt(st);
+            VG_(printf)("\n");
+            VG_(tool_panic)("addrcheck: unhandled IRStmt");
+      }
+
+      /* If needed, add a helper call. */
+      if (aexpr) {
+         tl_assert(hsz > 0);
+         switch (hsz) {
+            case 4:
+               if (isLoad)
+                  di = unsafeIRDirty_0_N( 1, "ach_LOAD4", &ach_LOAD4,
+                                          mkIRExprVec_1(aexpr));
+               else
+                  di = unsafeIRDirty_0_N( 1, "ach_STORE4", &ach_STORE4,
+                                          mkIRExprVec_1(aexpr));
+               break;
+            case 2:
+               if (isLoad)
+                  di = unsafeIRDirty_0_N( 1, "ach_LOAD2", &ach_LOAD2,
+                                          mkIRExprVec_1(aexpr));
+               else
+                  di = unsafeIRDirty_0_N( 1, "ach_STORE2", &ach_STORE2,
+                                          mkIRExprVec_1(aexpr));
+               break;
+            case 1:
+               if (isLoad)
+                  di = unsafeIRDirty_0_N( 1, "ach_LOAD1", &ach_LOAD1,
+                                          mkIRExprVec_1(aexpr));
+               else
+                  di = unsafeIRDirty_0_N( 1, "ach_STORE1", &ach_STORE1,
+                                          mkIRExprVec_1(aexpr));
+               break;
+            default:
+               if (isLoad)
+                  di = unsafeIRDirty_0_N( 
+                          2, "ach_LOADN", &ach_LOADN,
+                          mkIRExprVec_2(aexpr,mkIRExpr_HWord(hsz)));
+               else
+                  di = unsafeIRDirty_0_N( 
+                          2, "ach_STOREN", &ach_STOREN,
+                          mkIRExprVec_2(aexpr,mkIRExpr_HWord(hsz)));
+               break;
+	 }
+
+	 /* If the call has arisen as a result of a dirty helper which
+            references memory, we need to inherit the guard from the
+            dirty helper. */
+         if (guard)
+            di->guard = dopyIRExpr(guard);
+
+         /* emit the helper call */
+         addStmtToIRBB( bb, IRStmt_Dirty(di) );
+
+      }
+
+      /* And finally, copy the expr itself to the output. */
+      addStmtToIRBB( bb, st );
+   }
+
+   return bb;
 }
 
 
@@ -1123,7 +1158,7 @@ UCodeBlock* SK_(instrument)(UCodeBlock* cb_in, Addr orig_addr)
 static
 Bool ac_is_valid_64k_chunk ( UInt chunk_number )
 {
-   sk_assert(chunk_number >= 0 && chunk_number < PRIMARY_SIZE);
+   tl_assert(chunk_number >= 0 && chunk_number < PRIMARY_SIZE);
    if (primary_map[chunk_number] == DSM_NOTADDR) {
       /* Definitely not in use. */
       return False;
@@ -1139,7 +1174,7 @@ static
 Bool ac_is_valid_address ( Addr a )
 {
    UChar abits;
-   sk_assert(IS_ALIGNED4_ADDR(a));
+   tl_assert(VG_IS_4_ALIGNED(a));
    abits = get_abits4_ALIGNED(a);
    if (abits == VGM_NIBBLE_VALID) {
       return True;
@@ -1152,9 +1187,9 @@ Bool ac_is_valid_address ( Addr a )
 /* Leak detector for this tool.  We don't actually do anything, merely
    run the generic leak detector with suitable parameters for this
    tool. */
-static void ac_detect_memory_leaks ( LeakCheckMode mode )
+static void ac_detect_memory_leaks ( ThreadId tid, LeakCheckMode mode )
 {
-   MAC_(do_detect_memory_leaks) ( mode, ac_is_valid_64k_chunk, ac_is_valid_address );
+   MAC_(do_detect_memory_leaks) ( tid, mode, ac_is_valid_64k_chunk, ac_is_valid_address );
 }
 
 
@@ -1162,13 +1197,13 @@ static void ac_detect_memory_leaks ( LeakCheckMode mode )
    Sanity check machinery (permanently engaged).
    ------------------------------------------------------------------ */
 
-Bool SK_(cheap_sanity_check) ( void )
+static Bool ac_cheap_sanity_check ( void )
 {
    /* nothing useful we can rapidly check */
    return True;
 }
 
-Bool SK_(expensive_sanity_check) ( void )
+static Bool ac_expensive_sanity_check ( void )
 {
    Int i;
 
@@ -1192,13 +1227,13 @@ Bool SK_(expensive_sanity_check) ( void )
 /*--- Client requests                                      ---*/
 /*------------------------------------------------------------*/
 
-Bool SK_(handle_client_request) ( ThreadId tid, UWord* arg, UWord *ret )
+static Bool ac_handle_client_request ( ThreadId tid, UWord* arg, UWord *ret )
 {
 #define IGNORE(what)                                                    \
    do {                                                                 \
       if (moans-- > 0) {                                                \
          VG_(message)(Vg_UserMsg,                                       \
-            "Warning: Addrcheck: ignoring `%s' request.", what);     \
+            "Warning: Addrcheck: ignoring '%s' request.", what);     \
          VG_(message)(Vg_UserMsg,                                       \
             "   To honour this request, rerun with --tool=memcheck.");  \
       }                                                                 \
@@ -1207,18 +1242,18 @@ Bool SK_(handle_client_request) ( ThreadId tid, UWord* arg, UWord *ret )
    static Int moans = 3;
 
    /* Overload memcheck client reqs */
-   if (!VG_IS_SKIN_USERREQ('M','C',arg[0])
-    && VG_USERREQ__MALLOCLIKE_BLOCK != arg[0]
-    && VG_USERREQ__FREELIKE_BLOCK   != arg[0]
-    && VG_USERREQ__CREATE_MEMPOOL   != arg[0]
-    && VG_USERREQ__DESTROY_MEMPOOL  != arg[0]
-    && VG_USERREQ__MEMPOOL_ALLOC    != arg[0]
-    && VG_USERREQ__MEMPOOL_FREE     != arg[0])
+   if (!VG_IS_TOOL_USERREQ('M','C',arg[0])
+       && VG_USERREQ__MALLOCLIKE_BLOCK != arg[0]
+       && VG_USERREQ__FREELIKE_BLOCK   != arg[0]
+       && VG_USERREQ__CREATE_MEMPOOL   != arg[0]
+       && VG_USERREQ__DESTROY_MEMPOOL  != arg[0]
+       && VG_USERREQ__MEMPOOL_ALLOC    != arg[0]
+       && VG_USERREQ__MEMPOOL_FREE     != arg[0])
       return False;
 
    switch (arg[0]) {
       case VG_USERREQ__DO_LEAK_CHECK:
-         ac_detect_memory_leaks(arg[1] ? LC_Summary : LC_Full);
+         ac_detect_memory_leaks(tid, arg[1] ? LC_Summary : LC_Full);
 	 *ret = 0; /* return value is meaningless */
 	 break;
 
@@ -1261,27 +1296,46 @@ Bool SK_(handle_client_request) ( ThreadId tid, UWord* arg, UWord *ret )
 /*--- Setup                                                ---*/
 /*------------------------------------------------------------*/
 
-Bool SK_(process_cmd_line_option)(Char* arg)
+static Bool ac_process_cmd_line_option(Char* arg)
 {
    return MAC_(process_common_cmd_line_option)(arg);
 }
 
-void SK_(print_usage)(void)
+static void ac_print_usage(void)
 {  
    MAC_(print_common_usage)();
 }
 
-void SK_(print_debug_usage)(void)
+static void ac_print_debug_usage(void)
 {  
    MAC_(print_common_debug_usage)();
 }
-
+#endif
 
 /*------------------------------------------------------------*/
-/*--- Setup                                                ---*/
+/*--- Setup and finalisation                               ---*/
 /*------------------------------------------------------------*/
 
-void SK_(pre_clo_init)(void)
+// dummy instrument() function
+static IRBB* ac_instrument(IRBB* bb_in, VexGuestLayout* layout, 
+                           IRType gWordTy, IRType hWordTy )
+{
+   tl_assert(0);
+}
+
+static void ac_post_clo_init ( void )
+{
+}
+
+static void ac_fini ( Int exitcode )
+{
+   tl_assert(0);     // turn leak checking back on
+#if 0
+   MAC_(common_fini)( ac_detect_memory_leaks );
+#endif
+}
+
+static void ac_pre_clo_init(void)
 {
    VG_(details_name)            ("Addrcheck");
    VG_(details_version)         (NULL);
@@ -1291,14 +1345,49 @@ void SK_(pre_clo_init)(void)
    VG_(details_bug_reports_to)  (VG_BUGS_TO);
    VG_(details_avg_translation_sizeB) ( 135 );
 
+   VG_(basic_tool_funcs)          (ac_post_clo_init,
+                                   ac_instrument,
+                                   ac_fini);
+
+
+   VG_(printf)(
+"\n"
+"Addrcheck is currently not working, because:\n"
+" (a) it is not yet ready to handle the Vex IR and the use with 64-bit\n"
+"     platforms introduced in Valgrind 3.0.0\n"
+"\n"
+"Sorry for the inconvenience.  Let us know if this is a problem for you.\n");
+   VG_(exit)(1);
+
+
+#if 0
    VG_(needs_core_errors)         ();
-   VG_(needs_skin_errors)         ();
+   VG_(needs_tool_errors)         (MAC_(eq_Error),
+                                   ac_pp_Error,
+                                   MAC_(update_extra),
+                                   MAC_(shared_recognised_suppression),
+                                   MAC_(read_extra_suppression_info),
+                                   MAC_(error_matches_suppression),
+                                   MAC_(get_error_name),
+                                   MAC_(print_extra_suppression_info));
    VG_(needs_libc_freeres)        ();
-   VG_(needs_command_line_options)();
-   VG_(needs_client_requests)     ();
-   VG_(needs_syscall_wrapper)     ();
-   VG_(needs_sanity_checks)       ();
-   VG_(needs_shadow_memory)       ();
+   VG_(needs_command_line_options)(MAC_(process_common_cmd_line_option),
+                                   MAC_(print_common_usage),
+                                   MAC_(print_common_debug_usage));
+   VG_(needs_client_requests)     (ac_handle_client_request);
+   VG_(needs_sanity_checks)       (ac_cheap_sanity_check,
+                                   ac_expensive_sanity_check);
+
+   VG_(needs_malloc_replacement)  (MAC_(malloc),
+                                   MAC_(__builtin_new),
+                                   MAC_(__builtin_vec_new),
+                                   MAC_(memalign),
+                                   MAC_(calloc),
+                                   MAC_(free),
+                                   MAC_(__builtin_delete),
+                                   MAC_(__builtin_vec_delete),
+                                   MAC_(realloc),
+                                   MAC_MALLOC_REDZONE_SZB );
 
    MAC_( new_mem_heap)             = & ac_new_mem_heap;
    MAC_( ban_mem_heap)             = & ac_make_noaccess;
@@ -1306,58 +1395,50 @@ void SK_(pre_clo_init)(void)
    MAC_( die_mem_heap)             = & ac_make_noaccess;
    MAC_(check_noaccess)            = & ac_check_noaccess;
 
-   VG_(init_new_mem_startup)      ( & ac_new_mem_startup );
-   VG_(init_new_mem_stack_signal) ( & ac_make_accessible );
-   VG_(init_new_mem_brk)          ( & ac_make_accessible );
-   VG_(init_new_mem_mmap)         ( & ac_new_mem_mmap );
+   VG_(track_new_mem_startup)     ( & ac_new_mem_startup );
+   VG_(track_new_mem_stack_signal)( & ac_make_accessible );
+   VG_(track_new_mem_brk)         ( & ac_make_accessible );
+   VG_(track_new_mem_mmap)        ( & ac_new_mem_mmap );
    
-   VG_(init_copy_mem_remap)       ( & ac_copy_address_range_state );
+   VG_(track_copy_mem_remap)      ( & ac_copy_address_range_state );
       
-   VG_(init_die_mem_stack_signal) ( & ac_make_noaccess ); 
-   VG_(init_die_mem_brk)          ( & ac_make_noaccess );
-   VG_(init_die_mem_munmap)       ( & ac_make_noaccess ); 
+   VG_(track_die_mem_stack_signal)( & ac_make_noaccess ); 
+   VG_(track_die_mem_brk)         ( & ac_make_noaccess );
+   VG_(track_die_mem_munmap)      ( & ac_make_noaccess ); 
 
-   VG_(init_new_mem_stack_4)      ( & MAC_(new_mem_stack_4)  );
-   VG_(init_new_mem_stack_8)      ( & MAC_(new_mem_stack_8)  );
-   VG_(init_new_mem_stack_12)     ( & MAC_(new_mem_stack_12) );
-   VG_(init_new_mem_stack_16)     ( & MAC_(new_mem_stack_16) );
-   VG_(init_new_mem_stack_32)     ( & MAC_(new_mem_stack_32) );
-   VG_(init_new_mem_stack)        ( & MAC_(new_mem_stack)    );
+   VG_(track_new_mem_stack_4)     ( & MAC_(new_mem_stack_4)  );
+   VG_(track_new_mem_stack_8)     ( & MAC_(new_mem_stack_8)  );
+   VG_(track_new_mem_stack_12)    ( & MAC_(new_mem_stack_12) );
+   VG_(track_new_mem_stack_16)    ( & MAC_(new_mem_stack_16) );
+   VG_(track_new_mem_stack_32)    ( & MAC_(new_mem_stack_32) );
+   VG_(track_new_mem_stack)       ( & MAC_(new_mem_stack)    );
 
-   VG_(init_die_mem_stack_4)      ( & MAC_(die_mem_stack_4)  );
-   VG_(init_die_mem_stack_8)      ( & MAC_(die_mem_stack_8)  );
-   VG_(init_die_mem_stack_12)     ( & MAC_(die_mem_stack_12) );
-   VG_(init_die_mem_stack_16)     ( & MAC_(die_mem_stack_16) );
-   VG_(init_die_mem_stack_32)     ( & MAC_(die_mem_stack_32) );
-   VG_(init_die_mem_stack)        ( & MAC_(die_mem_stack)    );
+   VG_(track_die_mem_stack_4)     ( & MAC_(die_mem_stack_4)  );
+   VG_(track_die_mem_stack_8)     ( & MAC_(die_mem_stack_8)  );
+   VG_(track_die_mem_stack_12)    ( & MAC_(die_mem_stack_12) );
+   VG_(track_die_mem_stack_16)    ( & MAC_(die_mem_stack_16) );
+   VG_(track_die_mem_stack_32)    ( & MAC_(die_mem_stack_32) );
+   VG_(track_die_mem_stack)       ( & MAC_(die_mem_stack)    );
    
-   VG_(init_ban_mem_stack)        ( & ac_make_noaccess );
+   VG_(track_ban_mem_stack)       ( & ac_make_noaccess );
 
-   VG_(init_pre_mem_read)         ( & ac_check_is_readable );
-   VG_(init_pre_mem_read_asciiz)  ( & ac_check_is_readable_asciiz );
-   VG_(init_pre_mem_write)        ( & ac_check_is_writable );
-   VG_(init_post_mem_write)       ( & ac_make_accessible );
+   VG_(track_pre_mem_read)        ( & ac_check_is_readable );
+   VG_(track_pre_mem_read_asciiz) ( & ac_check_is_readable_asciiz );
+   VG_(track_pre_mem_write)       ( & ac_check_is_writable );
+   VG_(track_post_mem_write)      ( & ac_post_mem_write );
 
-   VGP_(register_profile_event) ( VgpSetMem,   "set-mem-perms" );
-   VGP_(register_profile_event) ( VgpCheckMem, "check-mem-perms" );
-   VGP_(register_profile_event) ( VgpESPAdj,   "adjust-ESP" );
+   VG_(register_profile_event) ( VgpSetMem,   "set-mem-perms" );
+   VG_(register_profile_event) ( VgpCheckMem, "check-mem-perms" );
+   VG_(register_profile_event) ( VgpESPAdj,   "adjust-ESP" );
 
    init_shadow_memory();
    MAC_(common_pre_clo_init)();
+#endif
 }
 
-void SK_(post_clo_init) ( void )
-{
-}
-
-void SK_(fini) ( Int exitcode )
-{
-   MAC_(common_fini)( ac_detect_memory_leaks );
-}
-
-VG_DETERMINE_INTERFACE_VERSION(SK_(pre_clo_init), 1./8)
+VG_DETERMINE_INTERFACE_VERSION(ac_pre_clo_init, 1./8)
 
 
 /*--------------------------------------------------------------------*/
-/*--- end                                                ac_main.c ---*/
+/*--- end                                                          ---*/
 /*--------------------------------------------------------------------*/
