@@ -33,7 +33,6 @@
 #include "pub_core_debuglog.h"    // For VG_(debugLog_vprintf)
 #include "pub_core_libcbase.h"
 #include "pub_core_libcassert.h"
-#include "pub_core_libcmman.h"
 #include "pub_core_libcprint.h"
 #include "pub_core_libcsignal.h"
 #include "pub_core_machine.h"
@@ -761,7 +760,9 @@ Char *VG_(describe_addr)(ThreadId tid, Addr addr)
 	 Bool keep = False;
 
 	 /* Add a new variable to the list */
-	 void newvar(Char *name, SymType *ty, Addr valuep, UInt size) {
+         // (the declaration avoids a compiler warning)
+	 //static void newvar(Char *name, SymType *ty, Addr valuep, UInt size);
+         void newvar(Char *name, SymType *ty, Addr valuep, UInt size) {
 	    Variable *v;
 
 	    /* have we been here before? */
@@ -981,7 +982,8 @@ Char *VG_(describe_addr)(ThreadId tid, Addr addr)
 	 Char expr[len*2];
 	 Char *sp = &expr[len];	/* pointer at start of string */
 	 Char *ep = sp;		/* pointer at end of string */
-	 void genstring(Variable *v, Variable *inner) {
+	 //  static void genstring(Variable *v, Variable *inner);  // avoid warning
+         void genstring(Variable *v, Variable *inner) {
 	    Variable *c = v->container;
 
 	    if (c != NULL)
